#include "ghostmp_mpi_util.h"

#include <cstddef>
#include <cassert>
#include <limits>
#include <string>
#include <vector>
#include <mpi.h>

namespace ghostmp
{

int send_tag(int dest, int tag, MPI_Comm comm)
{
  int dummy;
  return MPI_Send(&dummy, 0, MPI_INT, dest, tag, comm);
}

int recv_tag(int source, int tag, MPI_Comm comm, MPI_Status* status)
{
  int dummy;
  return MPI_Recv(&dummy, 0, MPI_INT, source, tag, comm, status);
}

const char* convert_thread_support_level_to_str(int support_level)
{
  switch (support_level)
  {
    case MPI_THREAD_SINGLE:
      return "MPI_THREAD_SINGLE";
    case MPI_THREAD_FUNNELED:
      return "MPI_THREAD_FUNNELED";
    case MPI_THREAD_SERIALIZED:
      return "MPI_THREAD_SERIALIZED";
    case MPI_THREAD_MULTIPLE:
      return "MPI_THREAD_MULTIPLE";
    default:
      return "UNKNOWN_THREAD_SUPPORT_LEVEL";
  }
}

int create_large_contiguous_datatype(std::size_t elem_size, std::size_t count,
    MPI_Datatype oldtype, MPI_Datatype* newtype)
{
  // this condition is needed to represent displacement later
  assert(count <= std::numeric_limits<MPI_Aint>::max() / elem_size);
  int block_length = std::numeric_limits<int>::max();
  assert(count / block_length <= static_cast<std::size_t>(std::numeric_limits<int>::max()));
  int q = count / block_length;
  int r = count % block_length;

  int rc;
  MPI_Datatype tmp_type;
  rc = MPI_Type_contiguous(q, oldtype, &tmp_type);
  if (rc != MPI_SUCCESS) { return rc; }
  rc = MPI_Type_commit(&tmp_type);
  if (rc != MPI_SUCCESS) { return rc; }

  int block_lengths[2] = {block_length, r};
  MPI_Aint disp = q * block_length * elem_size;
  MPI_Aint displacements[2] = {0, disp};
  MPI_Datatype datatypes[2] = {tmp_type, oldtype};

  rc = MPI_Type_create_struct(2, block_lengths, displacements, datatypes,
      newtype);
  if (rc != MPI_SUCCESS) { return rc; }
  rc = MPI_Type_free(&tmp_type);
  if (rc != MPI_SUCCESS) { return rc; }
  return MPI_SUCCESS;
}

int read_fasta(const char* filename, std::string& fasta, MPI_Comm comm)
{
  int rank;
  MPI_Comm_rank(comm, &rank);
  int nprocs;
  MPI_Comm_size(comm, &nprocs);

  if (nprocs == 1) {
    return 1;
  }

  std::vector<char> data;
  int rc = read_all(filename, data, comm);
  if (rc) { return rc; }

  if (nprocs == 2) { // in this case, the process (rank=1) will read whole fasta content and there are no oddments.
    fasta.assign(data.begin(), data.end());
    return 0;
  }

  std::size_t pos = data.size();
  std::vector<char> oddment;

  // detect oddments and send/recieve it.
  if (rank != 0) {
    //std::size_t pos = data.size();
    if (rank != nprocs - 1) {
      for (std::size_t i = (data.size() - 1) - 1; i > 0; --i) {
        if (data[i] == '\n' && data[i + 1] == '>') {
          pos = i + 1;
          break;
        }
      }
      if (pos == data.size()) { // cound not find "\n>"
        return 3;
      }
    }
    int dest = rank + 1;
    int source = rank - 1;
    int tag = 0;
    unsigned long send_size = data.size() - pos;
    unsigned long recv_size;
    //std::vector<char> oddment;
    if (rank == nprocs - 1) {
      MPI_Recv(&recv_size, 1, MPI_UNSIGNED_LONG, source, tag, comm, MPI_STATUS_IGNORE);
      oddment.resize(recv_size);
      MPI_Recv(&oddment[0], recv_size, MPI_CHAR, source, tag, comm, MPI_STATUS_IGNORE);
    } else if (rank == 1) {
      MPI_Send(&send_size, 1, MPI_UNSIGNED_LONG, dest, tag, comm);
      MPI_Send(&data[pos], send_size, MPI_CHAR, dest, tag, comm);
    } else {
      MPI_Sendrecv(&send_size, 1, MPI_UNSIGNED_LONG, dest, tag,
          &recv_size, 1, MPI_UNSIGNED_LONG, source, tag, comm, MPI_STATUS_IGNORE);
      oddment.resize(recv_size);
      MPI_Sendrecv(&data[pos], send_size, MPI_CHAR, dest, tag,
          &oddment[0], recv_size, MPI_CHAR, source, tag, comm, MPI_STATUS_IGNORE);
    }
    fasta.assign(oddment.begin(), oddment.end());
    fasta.insert(fasta.end(),data.begin(), data.begin() + pos);
  }

  return 0;
}

} // namespace ghostmp

